import argparse
import datetime
import logging
import time
import uuid
import os
import threading
from strenum import StrEnum
from dataclasses import dataclass, field
from typing import List, Optional, Tuple
from enum import Enum

from orthanc_api_client import OrthancApiClient, Study, Series, JobStatus, ResourceNotFound, InstancesSet, ResourceType, exceptions
from .helpers.scheduler import Scheduler
from .helpers.time_out import TimeOut
from .orthanc_monitor import OrthancMonitor, ChangeType

logger = logging.getLogger(__name__)


class ForwarderMode(StrEnum):
    DICOM = 'dicom'             # use DICOM
    DICOM_SERIES_BY_SERIES = 'dicom-series-by-series'             # use DICOM but create a new association for each series
    DICOM_WEB = 'dicom-web'     # use DicomWEB
    DICOM_WEB_SERIES_BY_SERIES = 'dicom-web-series-by-series'     # use DicomWEB but one request per series to avoid large payloads.  This will also split very large series into < 1GB requests
    PEERING = 'peering'         # use peering between 2 orthancs
    TRANSFER = 'transfer'       # use the transfer plugin accelerator between 2 orthancs


@dataclass
class ForwarderDestination:
    destination: str                        # the alias of the destination Modality, Peer or DicomWeb server
    forwarder_mode: ForwarderMode           # the mode to use to forward to the destination
    alternate_destination: str = None       # an alternate destination in case this one can not be contacted


# class ForwarderMetadata(Enum):
#     INSTANCE_PROCESSED = 4600
#     SENT_TO_DESTINATIONS = 4601
#     NEXT_RETRY = 4602


@dataclass
class ForwarderInstancesSetStatus:
    processed: bool = field(init=False, default=False)
    sent_to_destinations: List[str] = field(default_factory=list)
    retry_count: int = field(init=False, default=0)
    next_retry: Optional[datetime.datetime] = None


class OrthancForwarder:
    """
    Forwards everything Orthanc receives to another Orthanc peer, a DICOM modality or DicomWeb server.
    The Forwarder deletes the study/instances once they have been forwarded.

    The images may be modified before being sent.  In that case, you should:
    - either provide an instance_processor callback if you are modifying the instances 'in_place' (keeping the same Orthanc ids)
    - or override process() in a subclass
    The modifications shall be idempotent:  it shall always give the same result if you repeat the modification multiple times

    The images may be filtered out before being processed and forwarded.  In that case, you should:
    - either provide an instance_filter callback
    - or override filter() in a subclass
    Images that are filtered out are deleted from the forwarder.

    An OrthancForwarder may be triggered by two 'events': the stable study or the 'instance received' event.

    You might define multiple destinations -> the Forwarder will send the study to all destinations and delete the study only once the study has been sent to all destinations.
    i.e: destinations = [Destination(A, PEER), Destination(B, DICOM)]
    -> it will send to A and B

    You might also define alternate destinations that will be used when the primary destination is unreachable.
    i.e: destinations = [Destination(A, PEER, alternateDestination = Destination(B, DICOM))]
    -> it will try to send to A and, if A is down, will send to B

    If the forwarding fails, the Forwarder will retry to send the instances later on.

    The OrthancForwarder uses Orthanc metadata ranging between [4600, 4700[
    """

    retry_intervals = [60, 120, 300, 1800, 3600]

    def __init__(self,
                 source: OrthancApiClient,
                 destinations: List[ForwarderDestination],
                 trigger: ChangeType = ChangeType.STABLE_STUDY,
                 max_retry_count_at_startup: int = 5,
                 polling_interval_in_seconds: int = 1,
                 worker_threads_count: int = 1,
                 instance_filter = None,    # a method to filter instances.  Signature: Filter(api_client, instance_id) -> bool (returns True to keep an instance, returns False to delete it)
                 instance_processor = None  # a method to process instances before forwarding them.  Signature: Process(api_client, instance_id)
                 ):

        self._source = source
        self._destinations = destinations
        self._trigger = trigger
        self._max_retry_count_at_startup = max_retry_count_at_startup
        self._polling_interval_in_seconds = polling_interval_in_seconds
        self._is_running = False
        self._worker_threads_count = worker_threads_count
        self._execution_thread = None
        self._instance_filter = instance_filter
        self._instance_processor = instance_processor
        self._status = {}

    def wait_orthanc_started(self):
        retry = 0
        while not self._source.is_alive():
            logger.info("Waiting to connect to Orthanc")
            retry += 1
            if retry == self._max_retry_count_at_startup:
                logger.error("Could not connect to Orthanc at startup")
                raise Exception("Could not connect to Orthanc at startup")
            time.sleep(self._polling_interval_in_seconds)

        system = self._source.get_system()
        if "OverwriteInstances" not in system:
            logger.warning("Unable to check OverwriteInstances configuration")
        elif not system["OverwriteInstances"]:
            if self._instance_processor:
                logger.error("Orthanc Forwarder: when providing an instance_processor, you should have OverwriteInstances set to true to replace the instance with the new one")
                raise Exception("Invalid Orthanc configuration: OverwriteInstances is false")

    def execute(self):  # runs forever !
        self.wait_orthanc_started()

        while True:
            self.handle_all_content()
            time.sleep(self._polling_interval_in_seconds)

    def handle_all_content(self):
        if self._trigger == ChangeType.STABLE_STUDY:
            studies_ids = self._source.studies.get_all_ids()
            if len(studies_ids) > 0:
                for study_id in studies_ids:
                    self._handle_study(study_id=study_id,
                                       api_client=self._source)

        elif self._trigger == ChangeType.STABLE_SERIES:
            series_ids = self._source.series.get_all_ids()
            if len(series_ids) > 0:
                for id in series_ids:
                    self._handle_series(series_id=id,
                                        api_client=self._source)
            else:
                logger.warning(f"No series found in Orthanc at startup")

        elif self._trigger == ChangeType.NEW_INSTANCE:
            instances_ids = self._source.instances.get_all_ids()
            if len(instances_ids) > 0:
                for id in instances_ids:
                    self._handle_instance(instance_id=id,
                                          api_client=self._source)
            else:
                logger.warning(f"No series found in Orthanc at startup")
        else:
            raise NotImplementedError()


    def _thread_execute(self):
        while self._is_running:
            self.handle_all_content()
            time.sleep(self._polling_interval_in_seconds)

    def start(self):
        logger.info("Starting Orthanc Forwarder")

        # create execution thread
        self._execution_thread = threading.Thread(
            target = self._thread_execute,
            name = 'OrthancForwarder execution thread'
        )

        self.wait_orthanc_started()

        # start threads
        self._is_running = True
        self._execution_thread.start()

    def stop(self):
        logger.info("Stopping Orthanc Forwarder")

        self._is_running = False
        self._execution_thread.join()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()

    def _handle_study(self, study_id: str, api_client):
        instances_set = InstancesSet.from_study(api_client=api_client, study_id=study_id)
        self.handle_instances_set(instances_set)

    def _handle_series(self, series_id: str, api_client):
        instances_set = InstancesSet.from_series(api_client=api_client, series_id=series_id)
        self.handle_instances_set(instances_set)

    def _handle_instance(self, instance_id: str, api_client):
        instances_set = InstancesSet.from_instance(api_client=api_client, instance_id=instance_id)
        self.handle_instances_set(instances_set)

    def filter(self, instances_set: InstancesSet) -> InstancesSet:
        # this method can be overriden in a derived class.
        # By default, all instances not satisfying the filter are deleted
        if self._instance_filter:
            filtered = instances_set.filter_instances(self._instance_filter)
            logger.info(f"{instances_set} Deleting {len(filtered.instances_ids)} instances / {len(filtered.series_ids)} series that have been filtered out")
            filtered.delete()

        return instances_set

    def process(self, instances_set: InstancesSet) -> bool:
        # this method can be overriden in a derived class.

        if self._instance_processor:
            try:
                logger.info(f"{instances_set} Processing ...")

                instances_set.process_instances(self._instance_processor)

                logger.info(f"{instances_set} Processing ... done")
            except exceptions.OrthancApiException as ex:
                logger.error(f"{instances_set} Error while processing: {ex.msg}")
            except Exception as ex:
                logger.error(f"{instances_set} Error while processing: {ex}", exc_info=1)
                return False

        return True

    def forward(self, instances_set, already_sent_to_destinations: List[str]) -> List[str]:  # returns a list of destinations where the data has been sent
        sent_to_destinations = []

        # has_been_sent_to = self._status[instances_set.id].sent_to_destinations
        # check the metadata of a random instance to detect to which destinations it has already been sent (which would mean that we are retrying to process the set)
        #has_been_sent_to = self._source.instances.get_string_metadata(instances_set.instances_ids[0], metadata_name=str(ForwarderMetadata.SENT_TO_DESTINATIONS.value), default_value="").split(",")

        for dest in self._destinations:
            try:

                if dest.destination not in already_sent_to_destinations:
                    logger.info(f"{instances_set} Sending to {dest.destination} using {dest.forwarder_mode}")
                    self._forward_to_destination(
                        instances_set=instances_set,
                        destination=dest
                    )
                    logger.info(f"{instances_set} Sent")
                else:
                    logger.info(f"{instances_set} Sending ... already sent to {dest.destination} using {dest.forwarder_mode}")

                sent_to_destinations.append(dest.destination)
            except exceptions.OrthancApiException as ex:
                logger.error(f"{instances_set} Error while forwarding to {dest.destination}: {ex.msg}")
            except Exception as ex:
                logger.error(f"{instances_set} Error while forwarding to {dest.destination}: {ex}", exc_info=1)

        return sent_to_destinations
            # has_been_sent_to = self._source.instances.get_string_metadata(instances_set.instances_ids[0], metadata_name=str(ForwarderMetadata.SENT_TO_DESTINATIONS.value), default_value="").split(",")


        # only save the sent_to_destinations if there are multiple destinations and there has been a failure.  Otherwise, we'll delete the data anyway right after
        # if len(self._destinations) > 1 and len(sent_to_destinations) > 1:
        #     self._set_string_metadata(instances_set, metadata_name=str(ForwarderMetadata.SENT_TO_DESTINATIONS.value), content=",".join(sent_to_destinations))
        # self._status[instances_set.id].sent_to_destinations = sent_to_destinations


    def delete(self, instances_set):
        logger.info(f"{instances_set} Deleting ...")
        instances_set.delete()
        logger.info(f"{instances_set} Deleting ... Done")

    def handle_instances_set(self, instances_set: InstancesSet):

        if instances_set.id not in self._status:
            self._status[instances_set.id] = ForwarderInstancesSetStatus()
        else:  # this is a retry !
            if datetime.datetime.now() < self._status[instances_set.id].next_retry:
                logger.debug(f"{instances_set} Skipping while waiting for retry")
                return

        logger.info(f"{instances_set} Handling ...")

        # filter
        instances_set = self.filter(instances_set)

        # process
        if not self._status[instances_set.id].processed:
            self._status[instances_set.id].processed = self.process(instances_set)
        else:
            logger.info(f"{instances_set} Skipping processing that has already been performed")

        # forward
        sent_to_destinations = self.forward(instances_set, self._status[instances_set.id].sent_to_destinations)
        if len(sent_to_destinations) == len(self._destinations):
            # delete
            self.delete(instances_set)
        else:
            self._status[instances_set.id].sent_to_destinations = sent_to_destinations

            retry_count = self._status[instances_set.id].retry_count
            next_retry = datetime.datetime.now() + datetime.timedelta(seconds=self.retry_intervals[min(retry_count, len(self.retry_intervals) - 1)])
            logger.info(f"{instances_set} Failed, will retry at {next_retry}")

            self._status[instances_set.id].next_retry = next_retry
            self._status[instances_set.id].retry_count = retry_count + 1
            return

        logger.info(f"{instances_set} Handling ... Done")

    def _forward_to_destination(self, instances_set: InstancesSet, destination: ForwarderDestination):
        if destination.forwarder_mode == ForwarderMode.DICOM:
            self._source.modalities.send(
                target_modality=destination.destination,
                resources_ids=instances_set.instances_ids
            )
        elif destination.forwarder_mode == ForwarderMode.DICOM_SERIES_BY_SERIES:
            for s in instances_set.series_ids:
                self._source.modalities.send(
                    target_modality=destination.destination,
                    resources_ids=instances_set.get_instances_ids(series_id=s),
                    synchronous=True
                )
        elif destination.forwarder_mode == ForwarderMode.DICOM_WEB:
            for s in instances_set.series_ids:
                self._source.dicomweb_servers.send(
                    target_server=destination.destination,
                    resources_ids=instances_set.get_instances_ids(series_id=s)
                )
        elif destination.forwarder_mode == ForwarderMode.DICOM_WEB_SERIES_BY_SERIES:
            for s in instances_set.series_ids:
                series = self._source.series.get(s)
                if series.statistics.uncompressed_size > 1*1024*1024*1024:
                    logger.info(f"{instances_set} A series is larger than 1 GB, sending instance by instance")
                    for i in instances_set.get_instances_ids(seris_id=s):
                        self._source.dicomweb_servers.send(
                            target_server=destination.destination,
                            resources_ids=[i]
                        )
                else:
                    self._source.dicomweb_servers.send(
                        target_server=destination.destination,
                        resources_ids=instances_set.get_instances_ids(series_id=s)
                    )
        elif destination.forwarder_mode == ForwarderMode.PEERING:
            self._source.peers.send(
                target_peer=destination.destination,
                resources_ids=instances_set.instances_ids
            )

        elif destination.forwarder_mode == ForwarderMode.TRANSFER:
            self._source.transfers.send(
                target_peer=destination.destination,
                resources_ids=instances_set.instances_ids,
                resource_type=ResourceType.INSTANCE
            )

        else:
            raise NotImplementedError

    def _set_string_metadata(self, instances_set: InstancesSet, metadata_name: str, content: str):
            instances_set.process_instances(lambda c, i: c.instances.set_string_metadata(
                orthanc_id=i,
                metadata_name=metadata_name,
                content=content
            ))



import os
import typing
import pydicom
from enum import Enum


class DicomElementType(Enum):
    MANDATORY = 1  # for dicom tags that must be there (type 1 or 1c) -> throw an exception if not present
    REQUIRED = 2  # for dicom tags that are mandatory but accepts null value (type 2 or 2c)
    OPTIONAL = 3  # for dicom tags that are not mandatory (type 3)


class DicomWorklistBuilder:

    def __init__(self, folder: str = None):
        self._folder = folder

    def get_folder(self):
        return self._folder

    # to override in derived class to customize the workilist before it is saved to disk
    def customize(self, ds: pydicom.dataset.FileDataset) -> pydicom.dataset.FileDataset:
        return ds

    def generate(self, values: typing.Dict[str, str], file_name: str = None) -> str:
        """

        :param values: a Dictionary object created from an HL7 message.  Keys of the dico shall match pydicom tag names (i.e: AccessionNumber, PatientID, ...)
        :param filename:
        :return: the filename created
        """
        assert self._folder is not None or file_name is not None, "Please always provide a folder when creating the builder or provide a filename each time you generate a worklist"

        # now, let's try to build a DWL out of this
        file_meta = pydicom.dataset.Dataset()
        file_meta.MediaStorageSOPClassUID = '1.2.276.0.7230010.3.1.0.1'  # shall we use 1.2.840.10008.5.1.4.31 ?
        file_meta.MediaStorageSOPInstanceUID = pydicom.uid.generate_uid()
        file_meta.ImplementationClassUID = '1.2.826.0.1.3680043.9.6676.1.0.0.1'  # 1.2.826.0.1.3680043.9.6676. is Osimis prefix
        file_meta.ImplementationVersionName = 'OSIMISHL7DWL'

        ds = pydicom.dataset.FileDataset(file_name, {}, file_meta = file_meta, preamble = b'\0' * 128)
        if not "SOPInstanceUID" in values:
            values["SOPInstanceUID"] = pydicom.uid.generate_uid()
        if not "StudyInstanceUID" in values:
            values["StudyInstanceUID"] = pydicom.uid.generate_uid()  # set a default StudyInstanceUID.  It might be overriden from the dwl object

        # clip patient address at 64 chars (one of the Avignon CT does not handle them)
        patient_address = values.get('PatientAddress')
        if patient_address and len(patient_address) > 64:
            patient_address = patient_address[:60] + "..."
            values['PatientAddress'] = patient_address

        for field_name, element_type in [('AccessionNumber', DicomElementType.REQUIRED),
                                        ('InstitutionName', DicomElementType.OPTIONAL),
                                        ('InstitutionAddress', DicomElementType.OPTIONAL),
                                        ('PatientID', DicomElementType.MANDATORY),
                                        ('PatientName', DicomElementType.MANDATORY),
                                        ('PatientMotherBirthName', DicomElementType.OPTIONAL),
                                        ('PatientAddress', DicomElementType.OPTIONAL),
                                        ('PatientBirthDate', DicomElementType.MANDATORY),
                                        ('PatientSex', DicomElementType.MANDATORY),
                                        ('SOPInstanceUID', DicomElementType.MANDATORY),
                                        ('StudyInstanceUID', DicomElementType.MANDATORY),
                                        ('RequestingPhysician', DicomElementType.REQUIRED),
                                        ('ReferringPhysicianName', DicomElementType.REQUIRED),
                                        ('RequestedProcedureDescription', DicomElementType.REQUIRED),
                                        ('RequestedProcedureID', DicomElementType.MANDATORY),
                                        ('SpecificCharacterSet', DicomElementType.MANDATORY)
                          ]:
            self._add_field(ds, values, field_name, element_type)

        ds.ReferencedStudySequence = pydicom.sequence.Sequence()
        ds.ReferencedPatientSequence = pydicom.sequence.Sequence()

        step = pydicom.dataset.Dataset()
        step.ScheduledProcedureStepDescription = values.get('RequestedProcedureDescription')
        for field_name, element_type in [('Modality', DicomElementType.REQUIRED),
                                        ('ScheduledProcedureStepStartDate', DicomElementType.OPTIONAL),
                                        ('ScheduledProcedureStepStartTime', DicomElementType.OPTIONAL),
                                        ('ReferringPhysicianName', DicomElementType.REQUIRED),
                                        ('ScheduledStationAETitle', DicomElementType.MANDATORY),
                                        ('ScheduledPerformingPhysicianName', DicomElementType.REQUIRED),
                                        ('ScheduledProcedureStepID', DicomElementType.MANDATORY),
                                        ('ScheduledStationName', DicomElementType.REQUIRED),
                                        ]:
            self._add_field(step, values, field_name, element_type)

        ds.ScheduledProcedureStepSequence = pydicom.sequence.Sequence([step])

        ds = self.customize(ds)

        if file_name is None:  # if no filename provided, save in the folder
            file_name = os.path.join(self._folder, "{id}.wl".format(id = ds.AccessionNumber))

        ds.save_as(file_name, write_like_original = False)
        return file_name

    def _add_field(self, ds: pydicom.dataset.Dataset, values: typing.Dict[str, str], field_name: str, element_type: DicomElementType):
        if field_name in values:
            if values[field_name] is not None:
                ds.__setattr__(field_name, values.get(field_name))
            elif element_type == DicomElementType.REQUIRED:
                ds.__setattr__(field_name, '')
            elif element_type == DicomElementType.MANDATORY:
                raise Exception("missing field '{fieldName}'".format(fieldName = field_name))  # TODO: raise a dedicated exception
        elif element_type == DicomElementType.REQUIRED:
            ds.__setattr__(field_name, '')
        elif element_type == DicomElementType.MANDATORY:
            raise Exception("missing field '{fieldName}'".format(fieldName = field_name))  # TODO: raise a dedicated exception

# # Dicom-Meta-Information-Header
#
# # Dicom-Data-Set
# # Used TransferSyntax: Little Endian Explicit
# (0008,0005) CS [ISO_IR 100]                             #  10, 1 SpecificCharacterSet
# (0008,0018) UI [1.2.276.0.7230010.3.1.4.34260742.2908.1486551644.225000] #  56, 1 SOPInstanceUID
# (0008,0050) SH [63]                                     #   2, 1 AccessionNumber
# (0008,0080) LO [REMOVED]              #  26, 1 InstitutionName
# (0008,0081) ST [REMOVED]         #  30, 1 InstitutionAddress
# (0010,0010) PN [SURNAME^NAME]                          #  14, 1 PatientName
# (0010,0020) LO [38]                                     #   2, 1 PatientID
# (0010,0030) DA [19710711]                               #   8, 1 PatientBirthDate
# (0010,0040) CS [F]                                      #   2, 1 PatientSex
# (0020,000d) UI [1.2.276.0.7230010.3.1.2.34260742.2908.1486551644.224998] #  56, 1 StudyInstanceUID
# (0020,000e) UI [1.2.276.0.7230010.3.1.3.34260742.2908.1486551644.224999] #  56, 1 SeriesInstanceUID
# (0032,1032) PN (no value available)                     #   0, 0 RequestingPhysician
# (0032,1060) LO [TC BACINO]                              #  10, 1 RequestedProcedureDescription
# (0038,0010) LO [30]                                     #   2, 1 AdmissionID
# (0040,0100) SQ (Sequence with explicit length #=1)      # 108, 1 ScheduledProcedureStepSequence
#   (fffe,e000) na (Item with explicit length #=8)          # 100, 1 Item
#     (0008,0060) CS [CT]                                     #   2, 1 Modality
#     (0040,0001) AE [CT99]                                   #   4, 1 ScheduledStationAETitle
#     (0040,0002) DA [20170208]                               #   8, 1 ScheduledProcedureStepStartDate
#     (0040,0003) TM [083000]                                 #   6, 1 ScheduledProcedureStepStartTime
#     (0040,0006) PN (no value available)                     #   0, 0 ScheduledPerformingPhysicianName
#     (0040,0007) LO [TC BACINO]                              #  10, 1 ScheduledProcedureStepDescription
#     (0040,0009) SH [134]                                    #   4, 1 ScheduledProcedureStepID
#     (0040,0010) SH [TC]                                     #   2, 1 ScheduledStationName
#   (fffe,e00d) na (ItemDelimitationItem for re-encoding)   #   0, 0 ItemDelimitationItem
# (fffe,e0dd) na (SequenceDelimitationItem for re-encod.) #   0, 0 SequenceDelimitationItem
# (0040,1001) SH [134]                                    #   4, 1 RequestedProcedureID
# (0040,1003) SH [LOW]                                    #   4, 1 RequestedProcedurePriority



import os
import logging
import argparse
import datetime
from .helpers.scheduler import Scheduler
from orthanc_api_client import helpers
import logging
from orthanc_api_client import OrthancApiClient

logger = logging.getLogger(__name__)

class OrthancComparator:

    def __init__(self,
                 api_client: OrthancApiClient,
                 from_study_date: datetime.date,  # Start date
                 to_study_date: datetime.date,  # End date
                 modality: str,  # modality as configured in Orthanc (alias)
                 level: str = 'Series',
                 scheduler: Scheduler = None,
                 transfer_missing_to_modality: bool = False,
                 ignore_missing_from_orthanc: bool = False,
                 retrieve_missing_from_orthanc: bool = False,
                 ignore_missing_on_modality: bool = False
                 ):

        if level not in ["Study", "Series", "Instance"]:
            raise RuntimeError("Invalid value for argument 'level'")

        self._api_client = api_client
        self._modality = modality
        self._from_study_date = from_study_date
        self._to_study_date = to_study_date
        self._scheduler = scheduler
        self._level = level
        self._transfer_missing_to_modality = transfer_missing_to_modality
        self._ignore_missing_from_orthanc = ignore_missing_from_orthanc
        self._retrieve_missing_from_orthanc = retrieve_missing_from_orthanc
        self._ignore_missing_on_modality = ignore_missing_on_modality

    def execute(self):

        if self._from_study_date < self._to_study_date:
            # ex 20220101 -> 20220131
            inc_date = 1
            to_date = self._to_study_date + datetime.timedelta(days=1)
        else:
            # ex 20220131 -> 20220101
            inc_date = -1
            to_date = self._to_study_date - datetime.timedelta(days=1)

        current_date = self._from_study_date

        while current_date != to_date:
            self.compare_date(current_date)
            current_date += datetime.timedelta(days=inc_date)

    def compare_date(self, current_date: datetime.date):
        if self._scheduler:
            self._scheduler.wait_right_time_to_run()

        try:
            logger.info("Processing date {date}".format(date=str(current_date)))

            local_studies = self._api_client.studies.find(
                query={
                    'StudyDate': helpers.to_dicom_date(current_date)
                })
            remote_studies = self._api_client.modalities.query_studies(
                from_modality=self._modality,
                query={
                    'StudyDate': helpers.to_dicom_date(current_date),
                    'PatientID': '',
                    'PatientName': '',
                    'StudyInstanceUID': '',
                    'StudyDescription': ''
                })

            logger.info(f"{str(current_date)}")
            logger.info(f"=======================================")
            if len(local_studies) != len(remote_studies):
                logger.warning(f"WARNING {str(current_date)}: {len(local_studies)} studies in Orthanc, {len(remote_studies)} studies in modality")
            else:
                logger.info(f"found {len(local_studies)} studies on both side")

            for local_study in local_studies:

                try:
                    remote_match = [r for r in remote_studies if r.dicom_id == local_study.dicom_id]
                    study_summary = f"{local_study.patient_main_dicom_tags.get('PatientID')} - {local_study.patient_main_dicom_tags.get('PatientName')} - {local_study.main_dicom_tags.get('StudyDescription')}"

                    if len(remote_match) == 0 and not self._ignore_missing_on_modality:
                        logger.warning(f"WARNING {str(current_date)}, study missing on modality: {study_summary}")
                        if self._transfer_missing_to_modality:
                            logger.warning(f"WARNING {str(current_date)}, transferring study to modality: {study_summary}")
                            self._api_client.modalities.store(
                                target_modality=self._modality,
                                resources_ids=local_study.orthanc_id,
                                synchronous=True
                            )
                    elif len(remote_match) > 1:
                        logger.warning(f"WARNING {str(current_date)}, study found multiple times on modality: {study_summary}")
                    elif len(remote_match) == 1:
                        if self._level in ['Series', 'Instance']:
                            self.compare_study(orthanc_id=local_study.orthanc_id, dicom_id=local_study.dicom_id, study_summary=study_summary)
                except Exception as ex:
                    logger.error(f"ERROR: {str(ex)}")

            if not self._ignore_missing_from_orthanc:
                for remote_study in remote_studies:
                    try:

                        local_match = [l for l in local_studies if l.dicom_id == remote_study.dicom_id]
                        if len(local_match) == 0:
                            logger.warning(f"WARNING {str(current_date)}, study missing from Orthanc: {remote_study.tags.get('PatientID')} - {remote_study.tags.get('PatientName')} - {remote_study.tags.get('StudyDescription')}")
                            if self._retrieve_missing_from_orthanc:
                                logger.warning(f"WARNING {str(current_date)}, retrieving missing study from Orthanc: {remote_study.tags.get('PatientID')} - {remote_study.tags.get('PatientName')} - {remote_study.tags.get('StudyDescription')}")
                                self._api_client.modalities.move_study(
                                    from_modality=self._modality,
                                    dicom_id=remote_study.dicom_id
                                )
                        elif len(local_match) > 1:
                            logger.warning(f"WARNING {str(current_date)}, study found multiple times on Orthanc: {remote_study.tags.get('PatientID')} - {remote_study.tags.get('PatientName')} - {remote_study.tags.get('StudyDescription')}")
                        # elif self._ignore_missing_on_modality: # in this case only, study comparison has not been performed above -> do it now
                    except Exception as ex:
                        logger.error(f"ERROR: {str(ex)}")

        except Exception as ex:
            logger.error(f"ERROR: {str(ex)}")


    def compare_study(self, orthanc_id: str, dicom_id: str, study_summary: str):
        if self._scheduler:
            self._scheduler.wait_right_time_to_run()

        try:

            local_series = self._api_client.get_json(f"/studies/{orthanc_id}/series?expand")

            remote_series = self._api_client.modalities.query_series(
                from_modality=self._modality,
                query={
                    'StudyInstanceUID': dicom_id,
                    'SeriesInstanceUID': '',
                    'SeriesDescription': ''
                })

            if len(local_series) != len(remote_series):
                logger.warning(f"WARNING STUDY {study_summary}: {len(local_series)} series in Orthanc, {len(remote_series)} series in modality")

            for local_serie in local_series:
                local_dicom_id = local_serie.get('MainDicomTags').get('SeriesInstanceUID')
                remote_match = [r for r in remote_series if r.dicom_id == local_dicom_id]
                if len(remote_match) == 0 and not self._ignore_missing_on_modality:
                    logger.warning(f"WARNING STUDY {study_summary}, series missing from modality: {local_dicom_id}")
                    if self._transfer_missing_to_modality:
                        logger.warning(f"WARNING STUDY {study_summary}, transferring series to modality: {local_dicom_id}")
                        self._api_client.modalities.store(
                            target_modality=self._modality,
                            resources_ids=local_serie.get('ID'),
                            synchronous=True
                        )
                elif len(remote_match) > 1:
                    logger.warning(f"WARNING STUDY {study_summary}, series found multiple times on modality: {local_dicom_id}")
                elif len(remote_match) == 1:
                    series_summary = f"{local_dicom_id} (from STUDY {study_summary})"
                    if self._level in ['Instance']:
                        self.compare_series(
                            orthanc_id=local_serie.get('ID'),
                            dicom_id=local_dicom_id,
                            study_dicom_id=dicom_id,
                            series_summary=series_summary)

            if not self._ignore_missing_from_orthanc:
                for remote_serie in remote_series:
                    local_match = [l for l in local_series if l.get('MainDicomTags').get('SeriesInstanceUID') == remote_serie.dicom_id]
                    if len(local_match) == 0:
                        logger.warning(f"WARNING STUDY {dicom_id}, series missing from Orthanc: {remote_serie.dicom_id}")
                        if self._retrieve_missing_from_orthanc:
                            logger.warning(f"WARNING STUDY {dicom_id}, retrieving missing series from Orthanc: {remote_serie.dicom_id}")
                            self._api_client.modalities.move_series(
                                from_modality=self._modality,
                                dicom_id=remote_serie.dicom_id,
                                study_dicom_id=dicom_id
                            )
                    elif len(local_match) > 1:
                        logger.warning(f"WARNING STUDY {dicom_id}, series found multiple times on Orthanc: {remote_serie.dicom_id}")
        except Exception as ex:
            logger.exception(f"ERROR: {str(ex)}")


    def compare_series(self, orthanc_id: str, dicom_id: str, study_dicom_id: str, series_summary: str):

        try:
            local_instances = self._api_client.get_json(f"/series/{orthanc_id}/instances?expand")

            remote_instances = self._api_client.modalities.query_instances(
                from_modality=self._modality,
                query={
                    'SeriesInstanceUID': dicom_id,
                    'SOPInstanceUID': ''
                })

            if len(local_instances) != len(remote_instances):
                logger.warning(f"WARNING SERIES {series_summary}: {len(local_instances)} instances in Orthanc, {len(remote_instances)} instances in modality")

            success_count = 0
            failure_count = 0

            for local_instance in local_instances:
                try:
                    local_dicom_id = local_instance.get('MainDicomTags').get('SOPInstanceUID')
                    remote_match = [r for r in remote_instances if r.dicom_id == local_dicom_id]

                    if len(remote_match) == 0 and not self._ignore_missing_on_modality:
                        logger.warning(f"WARNING SERIES {series_summary}, instance missing from modality: {local_dicom_id}")
                        if self._transfer_missing_to_modality:
                            logger.warning(f"WARNING SERIES {series_summary}, transferring instance to modality: {local_dicom_id}")
                            self._api_client.modalities.store(
                                target_modality=self._modality,
                                resources_ids=local_instance.get('ID'),
                                synchronous=True
                            )
                            success_count += 1
                    elif len(remote_match) > 1:
                        logger.warning(f"WARNING SERIES {series_summary}, instance found multiple times on modality: {local_dicom_id}")
                except:
                    failure_count += 1

            if not self._ignore_missing_from_orthanc:
                for remote_instance in remote_instances:
                    try:
                        local_match = [l for l in local_instances if l.get('MainDicomTags').get('SOPInstanceUID') == remote_instance.dicom_id]
                        if len(local_match) == 0:
                            logger.warning(f"WARNING SERIES {series_summary}, instance missing from Orthanc: {remote_instance.dicom_id}")
                            if self._retrieve_missing_from_orthanc:
                                logger.warning(f"WARNING SERIES {series_summary}, retrieving instance missing from Orthanc: {remote_instance.dicom_id}")
                                self._api_client.modalities.move_instance(
                                    from_modality=self._modality,
                                    dicom_id=remote_instance.dicom_id,
                                    series_dicom_id=dicom_id,
                                    study_dicom_id=study_dicom_id
                                )
                                success_count += 1
                        elif len(local_match) > 1:
                            logger.warning(f"WARNING SERIES {series_summary}, instance found multiple times on Orthanc: {remote_instance.dicom_id}")
                    except Exception as ex:
                        failure_count += 1

            if failure_count > 0:
                logger.error(f"ERROR SERIES {series_summary}, transferring/retrieving instances: {failure_count} failure, {success_count} success")
            elif success_count > 0:
                logger.warning(f"WARNING SERIES {series_summary}, transferred: {success_count} instances")

        except Exception as ex:
            logger.error(f"ERROR: {str(ex)}")


if __name__ == '__main__':
    level = logging.INFO

    if os.environ.get('VERBOSE_ENABLED'):
        level = logging.DEBUG

    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    logging.getLogger("urllib3").setLevel(logging.WARNING)

    parser = argparse.ArgumentParser(description='Compare the content of Orthanc with the content of a remote modality')
    parser.add_argument('--url', type=str, default='http://localhost:8042', help='Orthanc url')
    parser.add_argument('--user', type=str, default=None, help='Orthanc user name')
    parser.add_argument('--password', type=str, default=None, help='Orthanc password')
    parser.add_argument('--modality', type=str, required=True, help='Alias of the modality to compare with (in Orthanc config)')
    parser.add_argument('--from_study_date', type=str, required=True, help='From Study Date (format 20190225)')
    parser.add_argument('--to_study_date', type=str, required=True, help='To Study Date (format 20190225)')
    parser.add_argument('--level', type=str, default='Series', help='Compare resources up to Study/Series/Instance level')
    parser.add_argument('--transfer_missing_to_modality', default=False, action='store_true', help='Transfer missing resources from Orthanc to the remote modality')
    parser.add_argument('--ignore_missing_from_orthanc', default=False, action='store_true', help="Don't generate a warning if resources are missing from Orthanc")
    parser.add_argument('--retrieve_missing_from_orthanc', default=False, action='store_true', help='Retrieve missing resources to Orthanc from the remote modality')
    parser.add_argument('--ignore_missing_on_modality', default=False, action='store_true', help="Don't generate a warning if resources are missing on the remote modality")
    # TODO parser.add_argument('--retrieve_missing_in_orthanc', default=False, action='store_true', help='Retrieve missing resources from remote modality into Orthanc')
    Scheduler.add_parser_arguments(parser)

    args = parser.parse_args()

    url = os.environ.get("ORTHANC_URL", args.url)
    user = os.environ.get("ORTHANC_USER", args.user)
    password = os.environ.get("ORTHANC_PWD", args.password)
    modality = os.environ.get("MODALITY", args.modality)
    level = os.environ.get("LEVEL", args.level)
    from_study_date = helpers.from_dicom_date(os.environ.get("FROM_STUDY_DATE", args.from_study_date))
    to_study_date = helpers.from_dicom_date(os.environ.get("TO_STUDY_DATE", args.to_study_date))
    transfer_missing_to_modality = os.environ.get("TRANSFER_MISSING_TO_MODALITY", args.transfer_missing_to_modality)
    ignore_missing_from_orthanc = os.environ.get("IGNORE_MISSING_FROM_ORTHANC", args.ignore_missing_from_orthanc)
    retrieve_missing_from_orthanc = os.environ.get("RETRIEVE_MISSING_FROM_ORTHANC", args.retrieve_missing_from_orthanc)
    ignore_missing_on_modality = os.environ.get("IGNORE_MISSING_ON_MODALITY", args.ignore_missing_on_modality)

    scheduler = Scheduler.create_from_args_and_env_var(args)

    comparator = OrthancComparator(
        api_client=OrthancApiClient(url, user=user, pwd=password),
        modality=modality,
        from_study_date=from_study_date,
        to_study_date=to_study_date,
        level=level,
        scheduler=scheduler,
        transfer_missing_to_modality=transfer_missing_to_modality,
        ignore_missing_from_orthanc=ignore_missing_from_orthanc,
        retrieve_missing_from_orthanc=retrieve_missing_from_orthanc,
        ignore_missing_on_modality=ignore_missing_on_modality
    )

    comparator.execute()






import argparse
import logging
import os
from strenum import StrEnum

from orthanc_api_client import OrthancApiClient, ResourceType, JobStatus, ResourceNotFound
from .helpers.scheduler import Scheduler
from .orthanc_monitor import OrthancMonitor, ChangeType

logger = logging.getLogger(__name__)


class ClonerMode(StrEnum):

    DEFAULT = 'Default'             # download instance and reupload them in new orthanc
    PEERING = 'Peering'             # use peering between 2 orthancs
    TRANSFER = 'Transfer'           # use the transfer plugin accelerator between 2 orthancs


class OrthancCloner(OrthancMonitor):

    def __init__(self,
                 source: OrthancApiClient,
                 destination: OrthancApiClient = None,          # must be defined for DEFAULT mode
                 worker_threads_count: int = 1,
                 persist_status_path: str = None,
                 polling_interval: float = 1,
                 mode: ClonerMode = ClonerMode.DEFAULT,
                 destination_peer: str = None,                    # the 'alias' of the destination peer if declared in Orthanc.  It must be defined for PEERING and TRANSFER mode
                 scheduler: Scheduler = None,
                 max_retries: int = 5,
                 error_folder_path: str = None
        ):
        super().__init__(
            api_client=source,
            worker_threads_count=worker_threads_count,
            persist_status_path=persist_status_path,
            polling_interval=polling_interval,
            scheduler=scheduler,
            max_retries=max_retries,
            error_folder_path=error_folder_path
        )

        self._destination = destination
        self._destination_peer = destination_peer
        self._mode = mode

        if self._scheduler:
            logger.info("Night & Week-end mode Enabled : " + str(self._scheduler._run_only_at_night_and_weekend))

        logger.info("Migrating with {n} threads".format(n=self._worker_threads_count))

        if self._mode in [ClonerMode.PEERING, ClonerMode.TRANSFER]:
            if destination_peer is None:
                raise ValueError("'destination_peer' must be defined in PEERING or TRANSFER mode")
        if self._mode == ClonerMode.DEFAULT:
            if destination is None:
                raise ValueError("'destination' must be defined in DEFAULT mode")

        if self._mode in [ClonerMode.PEERING, ClonerMode.DEFAULT]:
            self.add_handler(ChangeType.NEW_INSTANCE, self.handle_new_instance)
        elif self._mode == ClonerMode.TRANSFER:
            self.add_handler(ChangeType.STABLE_STUDY, self.handle_stable_study)


    def handle_new_instance(self, change_id, instance_id, api_client):
        try:
            if self._mode == ClonerMode.DEFAULT:
                dicom = api_client.instances.get_file(instance_id)

                self._destination.upload(dicom)
                logger.info(f"{change_id}, copied instance {instance_id}")
            elif self._mode == ClonerMode.PEERING:
                api_client.peers.send(target_peer=self._destination_peer, resources_ids=instance_id)

        except ResourceNotFound as ex:
            # An instance may have been deleted (by a user, a script,...) between the moment we called the /changes route
            # and the moment we try to handle it. So, the `get_file` will return a 404 http error.
            # In this case, simply log the error and do not retry in the OrthancMonitor (don't raise the Exception)
            if self._error_folder_path:
                error_file_path = os.path.join(self._error_folder_path, f"{change_id:010d}.NewInstance.error.txt")
                try:
                    with open(error_file_path, "wt") as f:
                        f.write(f"Error while cloning instance {instance_id}: {str(ex)}")
                except OSError as ex:
                    raise Exception(f"Could not write error report to file \"{ex.filename}\": {ex.strerror}")

        except Exception as ex:
            raise Exception(f"Error while cloning instance {instance_id}: {str(ex)}")

    def handle_stable_study(self, change_id, study_id, api_client):
        try:
            if self._mode == ClonerMode.TRANSFER:
                transfer_job = api_client.transfers.send_async(
                    target_peer=self._destination_peer,
                    resources_ids=study_id,
                    resource_type=ResourceType.STUDY
                )
                transfer_job.wait_completed(timeout=None)
                
                if transfer_job.info.status != JobStatus.SUCCESS:
                    raise Exception(str(transfer_job.info.content))

            logger.info(f"{change_id} transfered study {study_id}")

        except Exception as ex:
            raise Exception(f"Error while transferring study {study_id}: {str(ex)}")

# examples:
# python orthanc_tools/orthanc_cloner.py --source_url=http://192.168.0.10:8042 --source_user=user --source_pwd=pwd --dest_url=http://192.168.0.10:8042 --dest_user=user --dest_pwd=pwd
# python orthanc_tools/orthanc_cloner.py --source_url=http://192.168.0.10:8042 --source_user=user --source_pwd=pwd --dest_peer=pacs  --mode=Transfer

if __name__ == '__main__':
    level = logging.INFO

    if os.environ.get('VERBOSE_ENABLED'):
        level = logging.DEBUG

    logging.basicConfig(level=level, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    parser = argparse.ArgumentParser(description='Clone the content of an Orthanc into another Orthanc')
    parser.add_argument('--source_url', type=str, default=None, help='Orthanc source url')
    parser.add_argument('--source_user', type=str, default=None, help='Orthanc source user name')
    parser.add_argument('--source_pwd', type=str, default=None, help='Orthanc source password')
    parser.add_argument('--dest_url', type=str, default=None, help='Orthanc destination url')
    parser.add_argument('--dest_user', type=str, default=None, help='Orthanc destination user name')
    parser.add_argument('--dest_pwd', type=str, default=None, help='Orthanc destination password')
    parser.add_argument('--dest_peer', type=str, default=None, help='Orthanc destination peer (peer alias in source Orthanc)')
    parser.add_argument('--mode', type=str, default=None, help='Cloner Mode (Default, Peering, Transfer)')
    parser.add_argument('--persist_state_path', type=str, default=None, help='File path where the state of the cloner will be saved (to resume later)')
    parser.add_argument('--worker_threads_count', type=int, default=1, help='Number of worker threads')
    parser.add_argument('--error_folder_path', type=str, default=None, help='Folder path where to store error reports')
    parser.add_argument('--max_retries', type=int, default=5, help='Number of retries in case of error')

    Scheduler.add_parser_arguments(parser)

    args = parser.parse_args()

    source_url = os.environ.get("SOURCE_URL", args.source_url)
    source_user = os.environ.get("SOURCE_USER", args.source_user)
    source_pwd = os.environ.get("SOURCE_PWD", args.source_pwd)
    dest_url = os.environ.get("DEST_URL", args.dest_url)
    dest_user = os.environ.get("DEST_USER", args.dest_user)
    dest_pwd = os.environ.get("DEST_PWD", args.dest_pwd)
    dest_peer = os.environ.get("DEST_PEER", args.dest_peer)
    mode = os.environ.get("MODE", args.mode)
    persist_state_path = os.environ.get("PERSIST_STATE_PATH", args.persist_state_path)
    worker_threads_count = int(os.environ.get("WORKER_THREADS_COUNT", str(args.worker_threads_count)))
    error_folder_path = os.environ.get("ERROR_FOLDER_PATH", args.error_folder_path)
    max_retries = int(os.environ.get("MAX_RETRIES", args.max_retries))

    scheduler = Scheduler.create_from_args_and_env_var(args)

    destination = None
    if dest_url:
        destination = OrthancApiClient(dest_url, user=dest_user, pwd=dest_pwd)

    cloner = OrthancCloner(
        source=OrthancApiClient(source_url, user=source_user, pwd=source_pwd),
        destination=destination,
        persist_status_path=persist_state_path,
        mode=mode,
        destination_peer=dest_peer,
        scheduler=scheduler,
        worker_threads_count=worker_threads_count,
        error_folder_path=error_folder_path,
        max_retries=max_retries
    )

    cloner.execute(existing_changes_only=False)






import time
from .timer import Timer


class TimeOut(Timer):
    """
    Helper to checks if an elapsed time has been reached.
    """

    def __init__(self, timeout):
        """

        :param timeout: duration (in seconds) until the TimeOut object expires
        """
        self._timeout = timeout
        Timer.__init__(self)

    def is_expired(self):
        """
        Checks if the TimeOut has expired.

        example:
        # wait max 60 seconds until my system has started and check every seconds
        timeout = TimeOut(60)
        while not self._isStarted() and not timeout.is_expired():
            time.sleep(1)

        :return: True if expired, False otherwise
        """
        return self.get_elapsed_seconds() > self._timeout

    def wait_until_expired(self):
        """
        Waits until the TimeOut has expired.

        example:
        while True:
            timeout = TimeOut(self._pollingIntervalInSeconds)
            self._performLengthyOperation()
            timeout.waitExpired()
        """
        remaining = self._timeout - self.get_elapsed_seconds()
        if remaining >= 0:
            time.sleep(remaining)

    @staticmethod
    def wait_until_condition(condition, timeout, evaluate_interval = 1):
        """
        evaluate a condition (lambda) multiple times until it is statisfied or a timeout occurs
        :param condition: the lambda expression to evaluate
        :param timeout: the max time to wait (in seconds) until the condition is satisfied
        :param evaluate_interval: the interval (in seconds) between to evaluation of the condition
        :return: true if the condition is satisfed before the timeout expired, false otherwise
        """
        timeout = TimeOut(timeout)
        satisfied = False
        while not satisfied and not timeout.is_expired():
            satisfied = condition()
            if not satisfied:
                time.sleep(evaluate_interval)

        if not satisfied:  # if not satisfied after timeout, let's try a last time
            return condition()
        else:
            return True



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.md').read_text(encoding='utf-8')

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='orthanc_tools',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/guides/single-sourcing-package-version/
    version='0.8.3',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Python Orthanc Tools',  # Optional

    # This is an optional longer description of your project that represents
    # the body of text which users will see when they visit PyPI.
    #
    # Often, this is the same as your README, so you can just read it in from
    # that file directly (as we have already done above)
    #
    # This field corresponds to the "Description" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,  # Optional

    # Denotes that our long_description is in Markdown; valid values are
    # text/plain, text/x-rst, and text/markdown
    #
    # Optional if long_description is written in reStructuredText (rst) but
    # required for plain-text or Markdown; if unspecified, "applications should
    # attempt to render [the long_description] as text/x-rst; charset=UTF-8 and
    # fall back to text/plain if it is not valid rst" (see link below)
    #
    # This field corresponds to the "Description-Content-Type" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#description-content-type-optional
    long_description_content_type='text/markdown',  # Optional (see note above)

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/orthanc-team/python-orthanc-tools',  # Optional

    # This should be your name or the name of the organization which owns the
    # project.
    author='Orthanc Team',  # Optional

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email='info@orthanc.team',  # Optional

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',

        # Pick your license as you wish
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Programming Language :: Python :: 3.10",
        'Programming Language :: Python :: 3 :: Only',
    ],

    # This field adds keywords for your project which will appear on the
    # project page. What does your project relate to?
    #
    # Note that this is a list of additional keywords, separated
    # by commas, to be used to assist searching for the distribution in a
    # larger catalog.
    keywords='orthanc, dicom, rest api, hl7',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=["orthanc_tools", "orthanc_tools.hl7Lib", "orthanc_tools.helpers"],  # Required

    # Specify which Python versions you support. In contrast to the
    # 'Programming Language' classifiers above, 'pip install' will check this
    # and refuse to install the project if the version does not match. See
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='>=3.8, <4',

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/discussions/install-requires-vs-requirements/
    install_requires=[
        'requests',
        'orthanc-api-client>=0.11.8',
        'pydicom>=2.3.1',
        'hl7==0.4.2',
        'six'
    ],

    # List additional groups of dependencies here (e.g. development
    # dependencies). Users will be able to install these using the "extras"
    # syntax, for example:
    #
    #   $ pip install sampleproject[dev]
    #
    # Similar to `install_requires` above, these must be valid existing
    # projects.
    extras_require={  # Optional
        'dev': ['check-manifest', 'pytest'],
        'test': ['coverage', 'pytest'],
    },


    project_urls={  # Optional
        'Bug Reports': 'https://github.com/orthanc-team/python-orthanc-tools/issues',
        'Funding': 'https://orthanc.team',
        'Source': 'https://github.com/orthanc-team/python-orthanc-tools/',
    },
)



import os, argparse, sys
from .hl7_worklist_parser import Hl7WorklistParser
from .hl7_dicom_worklist_builder import DicomWorklistBuilder
import hl7, random
from datetime import datetime
import logging

logger = logging.getLogger(__name__)

#TODO: manage the automatic deletion of old files somewhere else


class Hl7OrmWorklistMsgHandler:

    def __init__(self,
                 parser: Hl7WorklistParser,
                 builder: DicomWorklistBuilder,
                 encoding: str = 'ascii'  # TODO: currently not used !
                 ):

        assert builder._folder is not None, "You must provide a DicomWorklistBuilder with a folder defined"
        logger.info("Creating ORM worklist message handler")

        self._parser = parser
        self._builder = builder

    def handle_orm_message(self, message: str) -> hl7.Message:

        # TODO: improve logging as it was done with osimis logger
        # with self._logger.context(str(self._messageCounter)):

        logger.info("received message:{eol}{message}".format(message = str(message).replace('\r', os.linesep), eol = os.linesep))
        hl7_request = hl7.parse(message)  # we need to parse it here only the build the response

        values = self._parser.parse(hl7_message = message)
        try:
            logger.info("generating file...")
            worklistFilePath = self._builder.generate(values)
        except Exception as e:
            logger.error("file not generated: {exception}".format(exception=e))
        logger.info("generated file: {path}".format(path = worklistFilePath))

        hl7_response = hl7.parse('MSH|^~\&|{sending_application}||{receiving_application}|{receiving_facility}|{date_time}||ACK^O01|{ack_message_id}|P|2.3||||||8859/1\rMSA|AA|{message_id}'.format(  # TODO: handle encoding
            sending_application = hl7_request['MSH.F5.R1.C1'],
            receiving_application = hl7_request['MSH.F3.R1.C1'],
            receiving_facility = hl7_request['MSH.F4.R1.C1'],
            date_time = datetime.now().strftime("%Y%m%d%H%M%S"),
            message_id = hl7_request['MSH.F10.R1.C1'],
            ack_message_id = str(random.randrange(0, 10**15))
        ))
        logger.info("sending response:{eol}{response}".format(response = str(hl7_response).replace('\r', os.linesep), eol = os.linesep))
        return hl7_response
